unit HeadList;

{ HeaderListBox -- Listbox component with integrated Header

  This component displays a list of items in columnar form. The items displayed
    are taken from the Strings property of the listbox in tab delimited form.

  Revision History
  09/26/96 Written by BA
  12/27/96 BA Added names to subcomponents for visibility in proxy system

  miscellaneous

    function MakeListString(Values: array of Variant): string;
      function for constructing a tab-delimited string for the items property --
      pass as many items as you have defined columns in the sections

  THeaderListBox

    Key Properties
      HeaderListBox contains most of the key properties of the ListBox and
        HeaderControl, and the following additional properties:

      BackgroundColor: TColor
      Color: TColor
      Cushion: Integer -- #of pixels between columns and text start/end
      Listbox: TListbox -- provided in case of need for access to members
      Header: THeaderControl -- provided in case of need for access to members
      HeaderHeight: integer  read GetHeaderHeight  write SetHeaderHeight default 17;
      Sections: THeaderSections -- the THeaderSection objects
      SelectedItems: Variant -- Variant array of Integer, or Null if none selected
      Values[vCol, vIndex: integer]: Variant -- read-only access to the column values

    Key Methods

    Key Events
      OnDrawCell: for drawing each section -- provides a flag for doing default drawing
      OnDrawSection: surfaced Header event
      OnSectionClick: surfaced Header event
    OnSectionResize: surfaced Header event
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, ExtCtrls, Menus;

type
  THeaderListbox = class;
  THLBDrawCellEvent = procedure(Sender: TObject; Col, Index: Integer;
    var Rect: TRect; var Text: string; State: TOwnerDrawState; var DoDefault: Boolean) of object;
  THLBSectionNotifyEvent = procedure (Sender: THeaderListbox;
     Section: THeaderSection) of object;
  THLBDrawSectionEvent = procedure (Sender: THeaderListbox;
     Section: THeaderSection; const Rect: TRect; Pressed: Boolean) of object;
  THLBCompareEvent = procedure (Sender: THeaderListbox;
     Column: Byte; Item1, Item2: Integer; var Compare: Integer) of object;

  THeaderListbox = class(TCustomPanel)
  private
    FHeader: THeaderControl;
    FListbox: TListbox;
    FCushion: Integer;
    FPopupMenu: TPopupMenu;
    FOnDrawCell: THLBDrawCellEvent;
    FOnClick: TNotifyEvent;
    FOnDblClick: TNotifyEvent;
    FOnMouseDown: TMouseEvent;
    FOnMouseUp: TMouseEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnDrawSection: THLBDrawSectionEvent;
    FOnSectionClick: THLBSectionNotifyEvent;
    FOnSectionResize: THLBSectionNotifyEvent;
    FOnCompare: THLBCompareEvent;
    LastPoint: Integer;
    function GetBackgroundColor: TColor;
    procedure SetBackgroundColor(value: TColor);
    function GetColor: TColor;
    procedure SetColor(value: TColor);
    procedure SetCushion(value: Integer);
    function GetHeaderHeight: integer;
    procedure SetHeaderHeight(value: integer);
    function GetValues(vCol, vIndex: integer): Variant;
    function GetStringValues(vCol, vIndex: integer): string;
    function MakeLocalPoint(Control: TControl; X, Y: integer): TPoint;
    function MakeChildPoint(Control: TControl; X, Y: integer): TPoint;

    function GetItems: TStrings;
    procedure SetItems(Value: TStrings);
    function GetExtSel: Boolean;
    procedure SetExtSel(Value: Boolean);
    function GetMultiSel: Boolean;
    procedure SetMultiSel(Value: Boolean);
    function GetItemHeight: integer;
    procedure SetItemHeight(Value: integer);
    function GetItemIndex: integer;
    procedure SetItemIndex(Value: integer);
    function GetIntHeight: Boolean;
    procedure SetIntHeight(Value: Boolean);
    procedure SetPopupMenu(Value: TPopupMenu);
    function GetSections: THeaderSections;
    procedure SetSections(Value: THeaderSections);
    function GetSelected(Index: integer): Boolean;
    procedure SetSelected(Index: integer; Value: Boolean);
    function GetSelectedItems: Variant;
    procedure SetSelectedItems(value: Variant);
    function GetSorted: Boolean;
    procedure SetSorted(Value: Boolean);
    function GetSelCount: integer;
    function GetTopIndex: integer;
    procedure SetTopIndex(Value: integer);
    // event handlers for header and listbox events
    procedure DrawTrackingLine(HeaderControl: THeaderControl;
        Section: THeaderSection; Width: Integer; State: TSectionTrackState);
    procedure DrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
                       State: TOwnerDrawState);
    procedure Child_MouseDown(Sender: TObject; Button: TMouseButton; Shift:
                        TShiftState; X, Y: Integer);
    procedure Child_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState;
                      X, Y: Integer);
    procedure Child_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Header_DrawSection(HeaderControl: THeaderControl;
      Section: THeaderSection; const Rect: TRect; Pressed: Boolean);
    procedure Header_SectionClick(HeaderControl: THeaderControl; Section: THeaderSection);
    procedure Header_SectionResize(HeaderControl: THeaderControl; Section: THeaderSection);
    procedure List_Click(Sender: TObject);
    procedure List_DblClick(Sender: TObject);
    procedure CMCtl3DChanged(var Message: TMessage); message CM_Ctl3dChanged;
  protected
    property Hint;
  public
    SortAscending: Boolean;
    SortColumn: Byte;
    procedure Invalidate; override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Clear;
    procedure GetSelectedAsStrings(Result: TStrings);
    function ItemAtPos(Pos: TPoint; Existing: Boolean): Integer;
    function ItemRect(Index: Integer): TRect;
    procedure SelectAll; {selects all items}
    procedure Sort(Column: Byte; Ascending: Boolean);

    property Listbox: TListbox  read FListbox;
    property Header: THeaderControl  read FHeader;
    property SelCount: integer read GetSelCount;
    property Selected[Index: Integer]: Boolean read GetSelected write SetSelected;
    property SelectedItems: Variant  read GetSelectedItems  write SetSelectedItems;
    property TopIndex: Integer read GetTopIndex write SetTopIndex;
    property Values[vCol, vIndex: integer]: Variant read GetValues;
    property StringValues[vCol, vIndex: integer]: string read GetStringValues;
  published
    property Align;
    property BackgroundColor: TColor  read GetBackgroundColor  write SetBackgroundColor;
    property BorderStyle;
    property Color: TColor  read GetColor  write SetColor;
    property Ctl3D; //! StH: Property case is important!
    property Cushion: Integer  read FCushion  write SetCushion default 2;
    property ExtendedSelect: Boolean read GetExtSel write SetExtSel;
    property Font;
    property IntegralHeight: Boolean read GetIntHeight write SetIntHeight;
    property ItemHeight: integer read GetItemHeight write SetItemHeight;
    property ItemIndex: integer read GetItemIndex write SetItemIndex;
    property Items: TStrings read GetItems write SetItems;
    property HeaderHeight: integer  read GetHeaderHeight  write SetHeaderHeight default 17;
    property MultiSelect: Boolean read GetMultiSel write SetMultiSel;
    property ParentCtl3D; //! StH: Property case is important!
    property ParentFont;
    property PopupMenu: TPopupMenu read FPopupMenu write SetPopupMenu;
    property Sections: THeaderSections read GetSections write SetSections;
    property Sorted: Boolean  read GetSorted write SetSorted;
    property TabStop;
    property TabOrder;

    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnDrawCell: THLBDrawCellEvent read FOnDrawCell write FOnDrawCell;
    property OnDrawSection: THLBDrawSectionEvent read FOnDrawSection write FOnDrawSection;
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    property OnCompare: THLBCompareEvent read FOnCompare write FOnCompare;
    property OnEnter;
    property OnExit;
    property OnKeyPress;
    property OnKeyDown;
    property OnKeyUp;
    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnSectionClick: THLBSectionNotifyEvent read FOnSectionClick write FOnSectionClick;
    property OnSectionResize: THLBSectionNotifyEvent read FOnSectionResize write FOnSectionResize;
  end;

const
  {  ASCII codes }
  TAB = #9;

function MakeListString(Values: array of Variant): string;

procedure Register;

implementation

function MakeListString(Values: array of Variant): string;
{ make a delimiter separated string from the values passed }
var
  i: integer;

  function VarToStr(const Value: Variant): string;
  var
    dt_temp: TDateTime;
  begin
    case VarType(Value) of
    varDate:  begin
      { if no date part, then it is a time, otherwise assume a date only,
         unless time is non-zero }
      if Value < 1  then
        Result := FormatDateTime(ShortTimeFormat, VarToDateTime(Value))
      else begin
        dt_temp := VarToDateTime(Value);
        if Frac(dt_temp) = 0.0 then
          Result := FormatDateTime(ShortDateFormat, dt_temp)
        else
          Result := FormatDateTime(ShortDateFormat + ' ' + ShortTimeFormat, dt_temp);
      end;
      end
    else
      Result := System.VarToStr(Value);
    end;
  end;

begin
  Result := VarToStr(Values[0]);
  for i := 1 to High(Values) do begin
    Result := Result + TAB + VarToStr(Values[i]);
  end;
end;


{ THeaderListBox }

constructor THeaderListbox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);                      // call inherited
  ControlStyle := ControlStyle {+ [csFramed]} - [csSetCaption]; // no caption
  BevelOuter := bvNone;
  BorderStyle := bsSingle;
  Width := 120;                                  // set bounds
  Height := 100;
  ParentColor := True;       //makes header background same as background
  FCushion := 2;
  SortColumn := MAXBYTE;

  FHeader := THeaderControl.Create(Self);        // create header
  with FHeader do begin                          // set header props
    Parent := Self;
    { following to make it pretty and draw-contained properly }
    Align := alTop;
    Name := 'HeaderControl';
    HeaderHeight := 17;  {the default}

    OnSectionTrack := DrawTrackingLine;          // hook header events
    OnMouseDown := Child_MouseDown;
    OnMouseUp := Child_MouseUp;
    OnMouseMove := Child_MouseMove;
    OnDrawSection := Header_DrawSection;
    OnSectionClick := Header_SectionClick;
    OnSectionResize := Header_SectionResize;
  end;

  FListbox := TListbox.Create(Self);             // create listbox
  with FListbox do begin                         // set listbox props
    Parent := Self;
    Align := alClient;
    Name := 'Listbox';
    ParentColor := False;
    ParentCtl3d := False;
    Ctl3d := False;
    BorderStyle := bsNone;
    Color := clWindow;
    Style := lbOwnerDrawFixed;
    OnDrawItem := Self.DrawItem;                      // hook listbox events
    OnClick := List_Click;
    OnDblClick := List_DblClick;
    OnMouseDown := Child_MouseDown;
    OnMouseUp := Child_MouseUp;
    OnMouseMove := Child_MouseMove;
  end;
end;

destructor THeaderListbox.Destroy;
begin
  //FHeader.Free;                                  // free header
  //FListbox.Free;                                 // free listbox
  inherited Destroy;                             // call inherited
end;

procedure THeaderListBox.Invalidate;
begin
  if Assigned(FHeader) then
    FHeader.Invalidate;
  if Assigned(FListBox) then
    FListbox.Invalidate;

  inherited Invalidate;
end;

procedure THeaderListBox.Header_DrawSection(HeaderControl: THeaderControl;
  Section: THeaderSection; const Rect: TRect; Pressed: Boolean);
begin
  {note: only called when Section style is owner draw }
  if Assigned(FOnDrawSection) then
    FOnDrawSection(Self, Section, Rect, Pressed);
end;

procedure THeaderListbox.DrawTrackingLine(HeaderControl: THeaderControl;
                         Section: THeaderSection; Width: Integer;
                         State: TSectionTrackState);
{ OnSectionTrack handler for header.  This procedure is called to }
{ draw the tracking line in the listbox as the  header is sized.  }
var
  ref_x: integer;
begin
  with FListbox.Canvas do begin
    Pen.Mode := pmNot;                   // use NOT pen mode
    if (LastPoint > 0) then begin        // if not first line...
      MoveTo(LastPoint, 0);
      LineTo(LastPoint, FListbox.ClientHeight);  // erase last line
    end;
    if State <> tsTrackEnd then begin            // if still tracking...
      ref_x := Section.Left + Width;
      MoveTo(ref_x, 0);
      LineTo(ref_x, FListbox.ClientHeight);  // draw new line
      LastPoint := ref_x;         // save line position
    end
    else begin                                   // otherwise...
      LastPoint := 0;                            // reset line position
      Invalidate;
    end;
  end;
end;

procedure THeaderListbox.CMCtl3DChanged(var Message: TMessage);
begin
  if BorderStyle = bsSingle then RecreateWnd;
  inherited;
end;

procedure THeaderListbox.DrawItem(Control: TWinControl; Index: Integer;
                                  Rect: TRect; State: TOwnerDrawState);
{ OnDrawItem handler for listbox. This procedure draws each delimited }
{ string into the listbox based on the position of the sections. }
var
 {Note: mark1 is a TAB or beginning of line; mark2 is a TAB or Length }
  i, len_text, scan_pos, mark1, mark2: integer;
  scan_text, cur_text: String;
  DoDefault: Boolean;

  procedure Scan;
  { scans string for next delimiter or end of string }
  begin
    while (scan_pos <= len_text) do begin
      Inc(scan_pos);
      {note: must evaluate first expression first}
      if (scan_pos > len_text) or (scan_text[scan_pos] = TAB) then break;
    end;
  end;

  procedure Next;
  { mark the next (or first) column of text }
  begin
    mark1 := scan_pos;
    Scan;
    mark2 := scan_pos;
  end;


const
  { DrawText alignment flags }
  Alignment: array[TAlignment] of word = (dt_Left, dt_Right, dt_Center);
begin
  scan_text := FListbox.Items[Index];
  len_text := Length(scan_text);
  scan_pos := 0;
  Listbox.Canvas.FillRect(Rect);      // empty draw rect
  
  { iterate over all sections in header }
  for i := 0 to FHeader.Sections.Count - 1 do begin
    { create a draw rect based on header section }
    Rect.Left := FHeader.Sections[i].Left + FCushion;
    Rect.Right := FHeader.Sections[i].Right - FCushion;

    { prepare the column text, and call into owner-draw if exists }
    Next;
    cur_text := Copy(scan_text, mark1 + 1, ((mark2 - mark1) - 1));
    DoDefault := True;
    if Assigned(FOnDrawCell) then
      FOnDrawCell(Self, i, Index, Rect, cur_text, State, DoDefault);
    { do default text drawing, unless suppressed by owner-draw routine
      NOTE: Rect or Text could have changed}
    if DoDefault then
      DrawText(Listbox.Canvas.Handle, PChar(cur_text), -1, Rect,
               dt_SingleLine or dt_VCenter or Alignment[Sections[i].Alignment])
  end;
end;

function THeaderListbox.GetStringValues(vCol, vIndex: integer): string;
{ retrieves the value in the requested column and item
  note: assumption of this routine is that access will primarily be
  of first column, so no caching of current item or column has been done }
var
 {Note: mark1 is a TAB or beginning of line; mark2 is a TAB or Length }
  i, len_text, scan_pos,mark1, mark2: integer;
  scan_text: String;

  procedure Scan;
  { scans string for next delimiter or end of string }
  begin
    while (scan_pos <= len_text) do begin
      Inc(scan_pos);
      {note: must evaluate first expression first}
      if (scan_pos > len_text) or (scan_text[scan_pos] = TAB) then break;
    end;
  end;

  procedure Next;
  { mark the next (or first) column of text }
  begin
    mark1 := scan_pos;
    Scan;
    mark2 := scan_pos;
  end;

begin
  scan_text := FListbox.Items[vIndex];
  len_text := Length(scan_text);
  scan_pos := 0;
  for i := 0 to vCol do Next;
  Result := Copy(scan_text, mark1 + 1, ((mark2 - mark1) - 1))
end;

function THeaderListbox.GetValues(vCol, vIndex: integer): Variant;
begin
  Result := GetStringValues(vCol, vIndex);
end;

function THeaderListbox.GetSelectedItems: Variant;
{ get list of selected items; returns -1 if none }

  procedure GetSelectedItemsMulti;
  type
    TIntArray = array[0..0] of LongInt;
    PIntArray = ^TIntArray;
  var
    item_count, new_item_count, i: integer;
    pArray: PIntArray;
  begin
    item_count := FListBox.SelCount;
    {allocate enough memory to get an array of indexes of selected items}
    GetMem(pArray, item_count * SizeOf(Longint));
    try
      {note: new_item_count should be the same as item_count}
      new_item_count := FListbox.Perform(LB_GETSELITEMS, item_count, Longint(pArray));
      if new_item_count > 0 then begin
        Result := VarArrayCreate([0, new_item_count - 1], varInteger);
        for i := 0 to new_item_count - 1 do begin
          Result[i] := pArray^[i];
        end;
      end;
    finally
      FreeMem(pArray);
    end;
  end;

begin
  {note: when MultiSelect enabled, default is 0, not -1, thus our test below}
  Result := -1;
  if FListBox.SelCount = 1 then
    Result := FListbox.ItemIndex
  else if FListBox.SelCount > 1 then
    GetSelectedItemsMulti;
end;

procedure THeaderListbox.GetSelectedAsStrings(Result: TStrings);
{ get list of selected items; returns empty if none }

  procedure GetSelectedItemsMulti;
  type
    TIntArray = array[0..0] of LongInt;
    PIntArray = ^TIntArray;
  var
    item_count, new_item_count, i: integer;
    pArray: PIntArray;
  begin
    item_count := FListBox.SelCount;
    {allocate enough memory to get an array of indexes of selected items}
    GetMem(pArray, item_count * SizeOf(Longint));
    try
      {note: new_item_count should be the same as item_count}
      new_item_count := FListbox.Perform(LB_GETSELITEMS, item_count, Longint(pArray));
      if new_item_count > 0 then begin
        for i := 0 to new_item_count - 1 do begin
          Result.Add(FListbox.Items[pArray^[i]]);
        end;
      end;
    finally
      FreeMem(pArray);
    end;
  end;

begin
  if (Result = nil) then exit;
  {note: when MultiSelect enabled, default is 0, not -1, thus our test below}
  Result.Clear;
  if FListBox.SelCount = 1 then
    Result.Add(FListbox.Items[FListbox.ItemIndex])
  else if FListBox.SelCount > 1 then
    GetSelectedItemsMulti;
end;

procedure THeaderListbox.SelectAll;
{selects all items}
begin
  if not MultiSelect then
    raise Exception.Create('Cannot select all: MultiSelect is False');
  FListbox.Perform(LB_SETSEL, -1{True}, -1{all});
end;

procedure THeaderListbox.Sort(Column: Byte; Ascending: Boolean);
var
  i: Integer;
  SortOrder: array[0..1024] of integer;
  Strings: TStringList;

  function CompareStr(Item1, Item2: Integer): Integer;
  begin
    if Assigned(FOnCompare) then
      FOnCompare(Self, Column, Item1, Item2, Result)
    else
      Result := CompareText(StringValues[Column, Item1], StringValues[Column, Item2]);
  end;

  procedure QuickSortByCol(L, R: Integer);
  var
    I, J: Integer;
    P, T: Integer;
  begin
    repeat
      I := L;
      J := R;
      P := SortOrder[(L + R) shr 1];
      repeat
        while CompareStr(SortOrder[I], P) < 0 do Inc(I);
        while CompareStr(SortOrder[J], P) > 0 do Dec(J);
        if I <= J then
        begin
          T := SortOrder[I];
          SortOrder[I] := SortOrder[J];
          SortOrder[J] := T;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then QuickSortByCol(L, J);
      L := I;
    until I >= R;
  end;

begin
  if FListBox.Items.Count > 1024 then
    raise Exception.Create('Too many items to sort!');
  if FListBox.Items.Count = 0 then
    Exit;
  FListBox.Items.BeginUpdate;
  try
    for i := 0 to FListBox.Items.Count - 1 do
      SortOrder[i] := i;
    QuickSortByCol(0, FListBox.Items.Count - 1);
    Strings := TStringList.Create;
    try
      if Ascending then
      begin
        for i := 0 to FListBox.Items.Count - 1 do
          Strings.Add(FListBox.Items[SortOrder[i]]);
      end
      else
      begin
        for i := FListBox.Items.Count - 1 downto 0 do
          Strings.Add(FListBox.Items[SortOrder[i]]);
      end;
      FListBox.Items.Assign(Strings);
    finally
      Strings.Free;
    end;
  finally
    FListBox.Items.EndUpdate;
  end;
end;

procedure THeaderListbox.SetSelectedItems(value: Variant);
{ set the selected items in the list -- deselect all if Null
  value can be a single value, or a VarArray of integer
  NOTE: only legal to set a single value if not in Multiselect mode}
var
  i: integer;

begin
  if MultiSelect then begin
    { remove previous selections }
    FListbox.Perform(LB_SETSEL, 0{false}, -1{all});

    { if an array of items, then select them all;
      otherwise select the single item, or ignore if -1 }
    if VarIsArray(value) then begin
      for i := VarArrayLowBound(value,1) to VarArrayHighBound(Value,1) do begin
        FListbox.Perform(LB_SETSEL, 1{true}, value);
      end;
      end
    else if VarType(value) = varInteger then
      if value <> -1 then
        FListbox.Perform(LB_SETSEL, 1{true}, value);
    end
  else begin
    {select the single item -- automatically deselects the previous item}
    if VarType(value) = varInteger then begin
      FListbox.ItemIndex := value;
    end;
  end;
end;


procedure THeaderListbox.SetCushion(value: Integer);
begin
  if FCushion = value then exit;
  FCushion := value;
  Invalidate;
end;

procedure THeaderListbox.List_Click(Sender: TObject);
begin
  if Assigned(FOnClick) then FOnClick(Self);
end;

procedure THeaderListbox.List_DblClick(Sender: TObject);
begin
  if Assigned(FOnDblClick) then
    FOnDblClick(Self);
end;

function THeaderListbox.MakeLocalPoint(Control: TControl;
                                       X, Y: integer): TPoint;
{ this function takes a coordinate relative to some arbitrary }
{ control and returns a coordinate relative to self }
begin
  Result.X := X;
  Result.Y := Y;
  Result := ScreenToClient(Control.ClientToScreen(Result));
end;

function THeaderListbox.MakeChildPoint(Control: TControl;
                                       X, Y: integer): TPoint;
{ this function takes a coordinate relative to self }
{ and returns a coordinate relative to some arbitrary control }
begin
  Result.X := X;
  Result.Y := Y;
  Result := Control.ScreenToClient(ClientToScreen(Result));
end;

procedure THeaderListbox.Child_MouseDown(Sender: TObject; Button: TMouseButton;
                                    Shift: TShiftState; X, Y: Integer);
{ OnMouseDown handler for header and listbox.  Resurfaces event. }
var
  Pt: TPoint;
begin
  Pt := MakeLocalPoint(Sender as TControl, X, Y);
  if Assigned(FOnMouseDown) then FOnMouseDown(Self, Button, Shift, Pt.X, Pt.Y);
end;

procedure THeaderListbox.Child_MouseUp(Sender: TObject; Button: TMouseButton; Shift:
                                 TShiftState; X, Y: Integer);
{ OnMouseUp handler for header and listbox.  Resurfaces event. }
var
  Pt: TPoint;
begin
  Pt := MakeLocalPoint(Sender as TControl, X, Y);
  if Assigned(FOnMouseUp) then FOnMouseUp(Self, Button, Shift, Pt.X, Pt.Y);
end;

procedure THeaderListbox.Child_MouseMove(Sender: TObject; Shift: TShiftState;
                                   X, Y: Integer);
{ OnMouseMove handler for header and listbox.  Resurfaces event. }
var
  Pt: TPoint;
begin
  Pt := MakeLocalPoint(Sender as TControl, X, Y);
  if Assigned(FOnMouseMove) then FOnMouseMove(Self, Shift, Pt.X, Pt.Y);
end;

procedure THeaderListbox.Header_SectionClick(HeaderControl: THeaderControl;
  Section: THeaderSection);
begin
  if Assigned(FOnSectionClick) then FOnSectionClick(Self, Section);
end;

procedure THeaderListbox.Header_SectionResize(HeaderControl: THeaderControl;
  Section: THeaderSection);
begin
  if Assigned(FOnSectionResize) then FOnSectionResize(Self, Section);
end;

procedure THeaderListbox.SetPopupMenu(Value: TPopupMenu);
begin
  if FPopupMenu = value then exit;
  FPopupMenu := Value;                    // set everyone's popup menu
  FListbox.PopupMenu := Value;
  FHeader.PopupMenu := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

{ What follows are bunches of header and listbox properties which are }
{ resurfaced in the THeaderListbox component. }

procedure THeaderListbox.Clear;
begin
  FListBox.Clear;
end;

function THeaderListbox.ItemAtPos(Pos: TPoint; Existing: Boolean): Integer;
begin
  Pos := MakeChildPoint(FListBox, Pos.X, Pos.Y);
  Result := FListBox.ItemAtPos(Pos, Existing);
end;

function THeaderListbox.ItemRect(Index: Integer): TRect;
begin
  Result := FListBox.ItemRect(Index);
  Result.TopLeft := MakeLocalPoint(FListBox, Result.Left, Result.Top);
  Result.BottomRight := MakeLocalPoint(FListBox, Result.Right, Result.Bottom);
end;

function THeaderListbox.GetBackgroundColor: TColor;
begin
  Result := inherited Color;
end;

procedure THeaderListbox.SetBackgroundColor(value: TColor);
begin
  inherited Color := value;
end;

function THeaderListbox.GetColor: TColor;
begin
  Result := FListbox.Color;
end;

procedure THeaderListbox.SetColor(value: TColor);
begin
  FListbox.Color := value;
end;

function THeaderListbox.GetHeaderHeight: integer;
begin
  Result := FHeader.Height;
end;

procedure THeaderListbox.SetHeaderHeight(value: integer);
begin
  FHeader.Height := value;
end;

function THeaderListbox.GetItems: TStrings;
begin
  Result := FListbox.Items;
end;

procedure THeaderListbox.SetItems(Value: TStrings);
begin
  FListbox.Items.Assign(Value);
end;

function THeaderListbox.GetSections: THeaderSections;
begin
  Result := FHeader.Sections
end;

procedure THeaderListbox.SetSections(Value: THeaderSections);
begin
  FHeader.Sections.Assign(Value);
  Invalidate;
end;

function THeaderListbox.GetExtSel: Boolean;
begin
  Result := FListbox.ExtendedSelect;
end;

procedure THeaderListbox.SetExtSel(Value: Boolean);
begin
  FListbox.ExtendedSelect := Value;
end;

function THeaderListbox.GetMultiSel: Boolean;
begin
  Result := FListbox.MultiSelect;
end;

procedure THeaderListbox.SetMultiSel(Value: Boolean);
begin
  FListbox.MultiSelect := Value;
end;

function THeaderListbox.GetItemHeight: integer;
begin
  Result := FListbox.ItemHeight;
end;

procedure THeaderListbox.SetItemHeight(Value: integer);
begin
  FListbox.ItemHeight := Value;
end;

function THeaderListbox.GetItemIndex: integer;
begin
  Result := FListbox.ItemIndex;
end;

procedure THeaderListbox.SetItemIndex(Value: integer);
begin
  FListbox.ItemIndex := Value;
end;

function THeaderListbox.GetIntHeight: Boolean;
begin
  Result := FListbox.IntegralHeight;
end;

procedure THeaderListbox.SetIntHeight(Value: Boolean);
begin
  FListbox.IntegralHeight := Value;
end;

function THeaderListbox.GetSelected(Index: integer): Boolean;
begin
  Result := FListbox.Selected[Index];
end;

procedure THeaderListbox.SetSelected(Index: integer; Value: Boolean);
begin
  FListbox.Selected[Index] := Value;
end;

function THeaderListbox.GetSorted: Boolean;
begin
  Result := FListbox.Sorted;
end;

procedure THeaderListbox.SetSorted(Value: Boolean);
begin
  FListbox.Sorted := Value;
end;

function THeaderListbox.GetSelCount: integer;
begin
  Result := FListbox.SelCount;
end;

function THeaderListbox.GetTopIndex: integer;
begin
  Result := FListbox.TopIndex;
end;

procedure THeaderListbox.SetTopIndex(Value: integer);
begin
  FListbox.TopIndex := Value;
end;

procedure Register;
begin
  RegisterComponents('GExperts', [THeaderListbox]);
end;

end.
